package javax.swing.text;

import java.awt.Shape;
import java.awt.Component;
import java.awt.font.TextAttribute;
import java.text.AttributedString;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.text.BreakIterator;
import java.awt.FontMetrics;
import sun.swing.SwingUtilities2;
import java.awt.Graphics;
import java.awt.Container;
import javax.swing.JComponent;

public class Utilities
{
    public Utilities() {
    }
    
    static JComponent getJComponent(View view) {
        if (view != null) {
            Container container = view.getContainer();
            if ((container instanceof JComponent)) {
                return (JComponent)container;
            }
        }
        return null;
    }
    
    public static final int drawTabbedText(Segment segment, int n, int n2, Graphics graphics, TabExpander tabExpander, int n3) {
        return drawTabbedText(null, segment, n, n2, graphics, tabExpander, n3);
    }
    
    static final int drawTabbedText(View view, Segment segment, int n, int n2, Graphics graphics, TabExpander tabExpander, int n3) {
        return drawTabbedText(view, segment, n, n2, graphics, tabExpander, n3, (int[])null);
    }
    
    static final int drawTabbedText(View view, Segment segment, int x, int y, Graphics graphics, TabExpander tabExpander, int n, int[] array) {
        JComponent jComponent = getJComponent(view);
        FontMetrics fontMetrics = SwingUtilities2.getFontMetrics(jComponent, graphics);
        int n2 = x;
        char[] array2 = segment.array;
        int i = segment.offset;
        int length = 0;
        int offset = segment.offset;
        int n3 = 0;
        int n4 = -1;
        int n5 = 0;
        int n6 = 0;
        if (array != null) {
            int n7 = -n + i;
            View parent;
            if (view != null && (parent = view.getParent()) != null) {
                n7 += parent.getStartOffset();
            }
            n3 = array[0];
            n4 = array[1] + n7;
            n5 = array[2] + n7;
            n6 = array[3] + n7;
        }
        int n8;
        for (n8 = segment.offset + segment.count, ; i < n8; ++i) {
            if (array2[i] == '\t' || ((n3 != 0 || i <= n4) && array2[i] == ' ' && n5 <= i && i <= n6)) {
                if (length > 0) {
                    n2 = SwingUtilities2.drawChars(jComponent, graphics, array2, offset, length, x, y);
                    length = 0;
                }
                offset = i + 1;
                if (array2[i] == '\t') {
                    if (tabExpander != null) {
                        n2 = (int)tabExpander.nextTabStop((float)n2, n + i - i);
                    }
                    else {
                        n2 += fontMetrics.charWidth(' ');
                    }
                }
                else if (array2[i] == ' ') {
                    n2 += fontMetrics.charWidth(' ') + n3;
                    if (i <= n4) {
                        ++n2;
                    }
                }
                x = n2;
            }
            else if (array2[i] == '\n' || array2[i] == '\r') {
                if (length > 0) {
                    n2 = SwingUtilities2.drawChars(jComponent, graphics, array2, offset, length, x, y);
                    length = 0;
                }
                offset = i + 1;
                x = n2;
            }
            else {
                ++length;
            }
        }
        if (length > 0) {
            n2 = SwingUtilities2.drawChars(jComponent, graphics, array2, offset, length, x, y);
        }
        return n2;
    }
    
    public static final int getTabbedTextWidth(Segment segment, FontMetrics fontMetrics, int n, TabExpander tabExpander, int n2) {
        return getTabbedTextWidth(null, segment, fontMetrics, n, tabExpander, n2, (int[])null);
    }
    
    static final int getTabbedTextWidth(View view, Segment segment, FontMetrics fontMetrics, int n, TabExpander tabExpander, int n2, int[] array) {
        char[] array2 = segment.array;
        int i = segment.offset;
        int n3 = segment.offset + segment.count;
        int len = 0;
        int n4 = 0;
        int n5 = -1;
        int n6 = 0;
        int n7 = 0;
        if (array != null) {
            int n8 = -n2 + i;
            View parent;
            if (view != null && (parent = view.getParent()) != null) {
                n8 += parent.getStartOffset();
            }
            n4 = array[0];
            n5 = array[1] + n8;
            n6 = array[2] + n8;
            n7 = array[3] + n8;
        }
        for (; i < n3; ++i) {
            if (array2[i] == '\t' || ((n4 != 0 || i <= n5) && array2[i] == ' ' && n6 <= i && i <= n7)) {
                n += fontMetrics.charsWidth(array2, i - len, len);
                len = 0;
                if (array2[i] == '\t') {
                    if (tabExpander != null) {
                        n = (int)tabExpander.nextTabStop((float)n, n2 + i - i);
                    }
                    else {
                        n += fontMetrics.charWidth(' ');
                    }
                }
                else if (array2[i] == ' ') {
                    n += fontMetrics.charWidth(' ') + n4;
                    if (i <= n5) {
                        ++n;
                    }
                }
            }
            else if (array2[i] == '\n') {
                n += fontMetrics.charsWidth(array2, i - len, len);
                len = 0;
            }
            else {
                ++len;
            }
        }
        return n + fontMetrics.charsWidth(array2, n3 - len, len) - n;
    }
    
    public static final int getTabbedTextOffset(Segment segment, FontMetrics fontMetrics, int n, int n2, TabExpander tabExpander, int n3) {
        return getTabbedTextOffset(segment, fontMetrics, n, n2, tabExpander, n3, true);
    }
    
    static final int getTabbedTextOffset(View view, Segment segment, FontMetrics fontMetrics, int n, int n2, TabExpander tabExpander, int n3, int[] array) {
        return getTabbedTextOffset(view, segment, fontMetrics, n, n2, tabExpander, n3, true, array);
    }
    
    public static final int getTabbedTextOffset(Segment segment, FontMetrics fontMetrics, int n, int n2, TabExpander tabExpander, int n3, boolean b) {
        return getTabbedTextOffset(null, segment, fontMetrics, n, n2, tabExpander, n3, b, (int[])null);
    }
    
    static final int getTabbedTextOffset(View view, Segment segment, FontMetrics fontMetrics, int n, int n2, TabExpander tabExpander, int n3, boolean b, int[] array) {
        if (n >= n2) {
            return 0;
        }
        char[] array2 = segment.array;
        int offset = segment.offset;
        int count = segment.count;
        int n4 = 0;
        int n5 = -1;
        int n6 = 0;
        int n7 = 0;
        if (array != null) {
            int n8 = -n3 + offset;
            View parent;
            if (view != null && (parent = view.getParent()) != null) {
                n8 += parent.getStartOffset();
            }
            n4 = array[0];
            n5 = array[1] + n8;
            n6 = array[2] + n8;
            n7 = array[3] + n8;
        }
        int n9 = segment.offset + segment.count;
        int i = segment.offset;
        while (i < n9) {
            if (array2[i] == '\t' || ((n4 != 0 || i <= n5) && array2[i] == ' ' && n6 <= i && i <= n7)) {
                if (array2[i] == '\t') {
                    if (tabExpander != null) {
                        n = (int)tabExpander.nextTabStop((float)n, n3 + i - offset);
                    }
                    else {
                        n += fontMetrics.charWidth(' ');
                    }
                }
                else if (array2[i] == ' ') {
                    n += fontMetrics.charWidth(' ') + n4;
                    if (i <= n5) {
                        ++n;
                    }
                }
            }
            else {
                n += fontMetrics.charWidth(array2[i]);
            }
            if (n2 < n) {
                int j;
                if (b) {
                    j = i + 1 - offset;
                    int charsWidth = fontMetrics.charsWidth(array2, offset, j);
                    int n10 = n2 - n;
                    if (n10 < charsWidth) {
                        while (j > 0) {
                            int n11 = (j > 1) ? fontMetrics.charsWidth(array2, offset, j - 1) : 0;
                            if (n10 >= n11) {
                                if (n10 - n11 < charsWidth - n10) {
                                    --j;
                                    break;
                                }
                                break;
                            }
                            else {
                                charsWidth = n11;
                                --j;
                            }
                        }
                    }
                }
                else {
                    for (j = i - offset; j > 0 && fontMetrics.charsWidth(array2, offset, j) > n2 - n; --j) {}
                }
                return j;
            }
            else {
                ++i;
            }
        }
        return count;
    }
    
    public static final int getBreakLocation(Segment text, FontMetrics fontMetrics, int n, int n2, TabExpander tabExpander, int n3) {
        char[] array = text.array;
        int offset = text.offset;
        int count = text.count;
        int tabbedTextOffset = getTabbedTextOffset(text, fontMetrics, n, n2, tabExpander, n3, false);
        if (tabbedTextOffset >= count - 1) {
            return count;
        }
        int i = offset + tabbedTextOffset;
        while (i >= offset) {
            int ch = array[i];
            if (ch < 'Ā') {
                if (Character.isWhitespace((char)ch)) {
                    tabbedTextOffset = i - offset + 1;
                    break;
                }
                --i;
            }
            else {
                BreakIterator lineInstance = BreakIterator.getLineInstance();
                lineInstance.setText(text);
                int preceding = lineInstance.preceding(i + 1);
                if (preceding > offset) {
                    tabbedTextOffset = preceding - offset;
                    break;
                }
                break;
            }
        }
        return tabbedTextOffset;
    }
    
    public static final int getRowStart(JTextComponent textComponent, int n) throws BadLocationException {
        Rectangle modelToView = textComponent.modelToView(n);
        if (modelToView == null) {
            return -1;
        }
        for (int y = modelToView.y; modelToView != null && y == modelToView.y; modelToView = ((--n >= 0) ? textComponent.modelToView(n) : null)) {
            if (modelToView.height != 0) {
            }
        }
        return n;
    }
    
    public static final int getRowEnd(JTextComponent textComponent, int n) throws BadLocationException {
        Rectangle modelToView = textComponent.modelToView(n);
        if (modelToView == null) {
            return -1;
        }
        int length = textComponent.getDocument().getLength();
        for (int y = modelToView.y; modelToView != null && y == modelToView.y; modelToView = ((++n <= length) ? textComponent.modelToView(n) : null)) {
            if (modelToView.height != 0) {
            }
        }
        return n;
    }
    
    public static final int getPositionAbove(JTextComponent textComponent, int n, int n2) throws BadLocationException {
        int n3 = getRowStart(textComponent, n) - 1;
        if (n3 < 0) {
            return -1;
        }
        int n4 = Integer.MAX_VALUE;
        int y = 0;
        Rectangle modelToView = null;
        if (n3 >= 0) {
            modelToView = textComponent.modelToView(n3);
            y = modelToView.y;
        }
        while (modelToView != null && y == modelToView.y) {
            int abs = Math.abs(modelToView.x - n2);
            if (abs < n4) {
                n = n3;
                n4 = abs;
            }
            modelToView = ((--n3 >= 0) ? textComponent.modelToView(n3) : null);
        }
        return n;
    }
    
    public static final int getPositionBelow(JTextComponent textComponent, int n, int n2) throws BadLocationException {
        int n3 = getRowEnd(textComponent, n) + 1;
        if (n3 <= 0) {
            return -1;
        }
        int n4 = Integer.MAX_VALUE;
        int length = textComponent.getDocument().getLength();
        int y = 0;
        Rectangle modelToView = null;
        if (n3 <= length) {
            modelToView = textComponent.modelToView(n3);
            y = modelToView.y;
        }
        while (modelToView != null && y == modelToView.y) {
            int abs = Math.abs(n2 - modelToView.x);
            if (abs < n4) {
                n = n3;
                n4 = abs;
            }
            modelToView = ((++n3 <= length) ? textComponent.modelToView(n3) : null);
        }
        return n;
    }
    
    public static final int getWordStart(JTextComponent textComponent, int i) throws BadLocationException {
        Document document = textComponent.getDocument();
        Element paragraphElement = getParagraphElement(textComponent, i);
        if (paragraphElement == null) {
            throw new BadLocationException("No word at " + i, i);
        }
        int startOffset = paragraphElement.getStartOffset();
        int min = Math.min(paragraphElement.getEndOffset(), document.getLength());
        Segment sharedSegment = SegmentCache.getSharedSegment();
        document.getText(startOffset, min - startOffset, sharedSegment);
        if (sharedSegment.count > 0) {
            BreakIterator wordInstance = BreakIterator.getWordInstance(textComponent.getLocale());
            wordInstance.setText(sharedSegment);
            int n = sharedSegment.offset + i - startOffset;
            if (n >= wordInstance.last()) {
                n = wordInstance.last() - 1;
            }
            wordInstance.following(n);
            i = startOffset + wordInstance.previous() - sharedSegment.offset;
        }
        SegmentCache.releaseSharedSegment(sharedSegment);
        return i;
    }
    
    public static final int getWordEnd(JTextComponent textComponent, int i) throws BadLocationException {
        Document document = textComponent.getDocument();
        Element paragraphElement = getParagraphElement(textComponent, i);
        if (paragraphElement == null) {
            throw new BadLocationException("No word at " + i, i);
        }
        int startOffset = paragraphElement.getStartOffset();
        int min = Math.min(paragraphElement.getEndOffset(), document.getLength());
        Segment sharedSegment = SegmentCache.getSharedSegment();
        document.getText(startOffset, min - startOffset, sharedSegment);
        if (sharedSegment.count > 0) {
            BreakIterator wordInstance = BreakIterator.getWordInstance(textComponent.getLocale());
            wordInstance.setText(sharedSegment);
            int n = i - startOffset + sharedSegment.offset;
            if (n >= wordInstance.last()) {
                n = wordInstance.last() - 1;
            }
            i = startOffset + wordInstance.following(n) - sharedSegment.offset;
        }
        SegmentCache.releaseSharedSegment(sharedSegment);
        return i;
    }
    
    public static final int getNextWord(JTextComponent textComponent, int endOffset) throws BadLocationException {
        Element element;
        int i;
        for (element = getParagraphElement(textComponent, endOffset), i = getNextWordInParagraph(textComponent, element, endOffset, false); i == -1; i = getNextWordInParagraph(textComponent, element, endOffset, true)) {
            endOffset = element.getEndOffset();
            element = getParagraphElement(textComponent, endOffset);
        }
        return i;
    }
    
    static int getNextWordInParagraph(JTextComponent textComponent, Element element, int n, boolean b) throws BadLocationException {
        if (element == null) {
            throw new BadLocationException("No more words", n);
        }
        Document document = element.getDocument();
        int startOffset = element.getStartOffset();
        int min = Math.min(element.getEndOffset(), document.getLength());
        if (n >= min || n < startOffset) {
            throw new BadLocationException("No more words", n);
        }
        Segment sharedSegment = SegmentCache.getSharedSegment();
        document.getText(startOffset, min - startOffset, sharedSegment);
        BreakIterator wordInstance = BreakIterator.getWordInstance(textComponent.getLocale());
        wordInstance.setText(sharedSegment);
        if (b && wordInstance.first() == sharedSegment.offset + n - startOffset && !Character.isWhitespace(sharedSegment.array[wordInstance.first()])) {
            return n;
        }
        int following = wordInstance.following(sharedSegment.offset + n - startOffset);
        if (following == -1 || following >= sharedSegment.offset + sharedSegment.count) {
            return -1;
        }
        if (!Character.isWhitespace(sharedSegment.array[following])) {
            return startOffset + following - sharedSegment.offset;
        }
        int next = wordInstance.next();
        if (next != -1) {
            n = startOffset + next - sharedSegment.offset;
            if (n != min) {
                return n;
            }
        }
        SegmentCache.releaseSharedSegment(sharedSegment);
        return -1;
    }
    
    public static final int getPreviousWord(JTextComponent textComponent, int n) throws BadLocationException {
        Element element;
        int i;
        for (element = getParagraphElement(textComponent, n), i = getPrevWordInParagraph(textComponent, element, n); i == -1; i = getPrevWordInParagraph(textComponent, element, n)) {
            n = element.getStartOffset() - 1;
            element = getParagraphElement(textComponent, n);
        }
        return i;
    }
    
    static int getPrevWordInParagraph(JTextComponent textComponent, Element element, int n) throws BadLocationException {
        if (element == null) {
            throw new BadLocationException("No more words", n);
        }
        Document document = element.getDocument();
        int startOffset = element.getStartOffset();
        int endOffset = element.getEndOffset();
        if (n > endOffset || n < startOffset) {
            throw new BadLocationException("No more words", n);
        }
        Segment sharedSegment = SegmentCache.getSharedSegment();
        document.getText(startOffset, endOffset - startOffset, sharedSegment);
        BreakIterator wordInstance = BreakIterator.getWordInstance(textComponent.getLocale());
        wordInstance.setText(sharedSegment);
        if (wordInstance.following(sharedSegment.offset + n - startOffset) == -1) {
            wordInstance.last();
        }
        int n2 = wordInstance.previous();
        if (n2 == sharedSegment.offset + n - startOffset) {
            n2 = wordInstance.previous();
        }
        if (n2 == -1) {
            return -1;
        }
        if (!Character.isWhitespace(sharedSegment.array[n2])) {
            return startOffset + n2 - sharedSegment.offset;
        }
        int previous = wordInstance.previous();
        if (previous != -1) {
            return startOffset + previous - sharedSegment.offset;
        }
        SegmentCache.releaseSharedSegment(sharedSegment);
        return -1;
    }
    
    public static final Element getParagraphElement(JTextComponent textComponent, int n) {
        Document document = textComponent.getDocument();
        if ((document instanceof StyledDocument)) {
            return ((StyledDocument)document).getParagraphElement(n);
        }
        Element defaultRootElement = document.getDefaultRootElement();
        Element element = defaultRootElement.getElement(defaultRootElement.getElementIndex(n));
        if (n >= element.getStartOffset() && n < element.getEndOffset()) {
            return element;
        }
        return null;
    }
    
    static boolean isComposedTextElement(Document document, int n) {
        Element element;
        for (element = document.getDefaultRootElement(); !element.isLeaf(); element = element.getElement(element.getElementIndex(n))) {}
        return isComposedTextElement(element);
    }
    
    static boolean isComposedTextElement(Element element) {
        return isComposedTextAttributeDefined(element.getAttributes());
    }
    
    static boolean isComposedTextAttributeDefined(AttributeSet set) {
        return set != null && set.isDefined(StyleConstants.ComposedTextAttribute);
    }
    
    static int drawComposedText(View view, AttributeSet set, Graphics graphics, int x, int y, int beginIndex, int endIndex) throws BadLocationException {
        Graphics2D g = (Graphics2D)graphics;
        AttributedString attributedString = (AttributedString)set.getAttribute(StyleConstants.ComposedTextAttribute);
        attributedString.addAttribute(TextAttribute.FONT, graphics.getFont());
        if (beginIndex >= endIndex) {
            return x;
        }
        return x + (int)SwingUtilities2.drawString(getJComponent(view), g, attributedString.getIterator(null, beginIndex, endIndex), x, y);
    }
    
    static void paintComposedText(Graphics graphics, Rectangle rectangle, GlyphView glyphView) {
        if ((graphics instanceof Graphics2D)) {
            Graphics2D g = (Graphics2D)graphics;
            int startOffset = glyphView.getStartOffset();
            int endOffset = glyphView.getEndOffset();
            AttributedString attributedString = (AttributedString)glyphView.getElement().getAttributes().getAttribute(StyleConstants.ComposedTextAttribute);
            int startOffset2 = glyphView.getElement().getStartOffset();
            int y = rectangle.y + rectangle.height - (int)glyphView.getGlyphPainter().getDescent(glyphView);
            int x = rectangle.x;
            attributedString.addAttribute(TextAttribute.FONT, glyphView.getFont());
            attributedString.addAttribute(TextAttribute.FOREGROUND, glyphView.getForeground());
            if (StyleConstants.isBold(glyphView.getAttributes())) {
                attributedString.addAttribute(TextAttribute.WEIGHT, TextAttribute.WEIGHT_BOLD);
            }
            if (StyleConstants.isItalic(glyphView.getAttributes())) {
                attributedString.addAttribute(TextAttribute.POSTURE, TextAttribute.POSTURE_OBLIQUE);
            }
            if (glyphView.isUnderline()) {
                attributedString.addAttribute(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
            }
            if (glyphView.isStrikeThrough()) {
                attributedString.addAttribute(TextAttribute.STRIKETHROUGH, TextAttribute.STRIKETHROUGH_ON);
            }
            if (glyphView.isSuperscript()) {
                attributedString.addAttribute(TextAttribute.SUPERSCRIPT, TextAttribute.SUPERSCRIPT_SUPER);
            }
            if (glyphView.isSubscript()) {
                attributedString.addAttribute(TextAttribute.SUPERSCRIPT, TextAttribute.SUPERSCRIPT_SUB);
            }
            SwingUtilities2.drawString(getJComponent(glyphView), g, attributedString.getIterator(null, startOffset - startOffset2, endOffset - startOffset2), x, y);
        }
    }
    
    static boolean isLeftToRight(Component component) {
        return component.getComponentOrientation().isLeftToRight();
    }
    
    static int getNextVisualPositionFrom(View view, int n, Position.Bias bias, Shape shape, int n2, Position.Bias[] array) throws BadLocationException {
        if (view.getViewCount() == 0) {
            return n;
        }
        int n3 = n2 == 1 || n2 == 7;
        int n6;
        if (n == -1) {
            int n8 = n3 ? (view.getViewCount() - 1) : 0;
            n6 = view.getView(n8).getNextVisualPositionFrom(n, bias, view.getChildAllocation(n8, shape), n2, array);
            if (n6 == -1 && n3 == 0 && view.getViewCount() > 1) {
                n6 = view.getView(1).getNextVisualPositionFrom(-1, array[0], view.getChildAllocation(1, shape), n2, array);
            }
        }
        else {
            int n4 = n3 ? -1 : 1;
            int n5;
            if (bias == Position.Bias.Backward && n > 0) {
                n5 = view.getViewIndex(n - 1, Position.Bias.Forward);
            }
            else {
                n5 = view.getViewIndex(n, Position.Bias.Forward);
            }
            View view2 = view.getView(n5);
            n6 = view2.getNextVisualPositionFrom(n, bias, view.getChildAllocation(n5, shape), n2, array);
            if ((n2 == 3 || n2 == 7) && view instanceof CompositeView && ((CompositeView)view).flipEastAndWestAtEnds(n, bias)) {
                n4 *= -1;
            }
            int n7 = n5 + n4;
            if (n6 == -1 && n7 >= 0 && n7 < view.getViewCount()) {
                n6 = view.getView(n7).getNextVisualPositionFrom(-1, bias, view.getChildAllocation(n7, shape), n2, array);
                if (n6 == n && array[0] != bias) {
                    return getNextVisualPositionFrom(view, n, array[0], shape, n2, array);
                }
            }
            else if (n6 != -1 && array[0] != bias && ((n4 == 1 && view2.getEndOffset() == n6) || (n4 == -1 && view2.getStartOffset() == n6)) && n7 >= 0 && n7 < view.getViewCount()) {
                View view3 = view.getView(n7);
                Shape childAllocation = view.getChildAllocation(n7, shape);
                Position.Bias bias2 = array[0];
                int nextVisualPosition = view3.getNextVisualPositionFrom(-1, bias, childAllocation, n2, array);
                if (array[0] == bias) {
                    n6 = nextVisualPosition;
                }
                else {
                    array[0] = bias2;
                }
            }
        }
        return n6;
    }
}
